from pylexers.RegularExpressions import (
    Concat,
    AtLeastOne,
    Or,
    Sigma,
    String,
    Star,
    Symbol,
    Epsilon,
    EmptySet,
)
from pylexers.NFALexer import NFALexer
from pylexers.DFALexer import DFALexer
from pylexers.DerivativeLexer import DerivativeLexer
from pylexers.BaseLexer import Lexer, Token, build_token_func



import dataclasses
from typing import Optional, Any


@dataclasses.dataclass
class SymbolTable:
    parent: Optional["SymbolTable"] = None
    identifiers: dict[str, Any] = dataclasses.field(default_factory=dict)
    user_higher_scope: bool = True

    def create_sub_scope(self):
        return SymbolTable(parent=self)

    def get_identifier(self, identifier):
        if self.does_identifier_exist(identifier):
            return self.identifiers[identifier][0]
        if self.parent and self.user_higher_scope:
            return self.parent.get_identifier(identifier)
        raise RuntimeError(f"Unknown identifier {identifier}")

    def get_identifier_type(self, identifier):
        if self.does_identifier_exist(identifier):
            return self.identifiers[identifier][1]
        if self.parent and self.user_higher_scope:
            return self.parent.get_identifier(identifier)
        raise SyntaxError(f"Unknown identifier {identifier}")

    def add_identifier(self, identifier: str, obj: Any, type: str):
        if self.does_identifier_exist(identifier):
            raise SyntaxError(f"Identifier already exists {identifier}")
        self.identifiers[identifier] = (obj, type)

    def reset_identifier(self, identifier: str, obj: Any, type: str):
        self.identifiers[identifier] = (obj, type)

    def remove_identifier(self, identifier: str):
        self.identifiers.pop(identifier)

    def does_identifier_exist(self, identifier: str):
        if identifier in self.identifiers:
            return True
        if self.parent and self.user_higher_scope:
            return self.parent.does_identifier_exist(identifier)
        return False



from pylexers.BaseLexer import Lexer
from pylexers.DFALexer.DFA import DFA
from pylexers.RegularExpressions.BaseRegularExpressions import _RegularExpression


class DFALexer(Lexer):
    def __init__(
        self, regular_expressions: list[_RegularExpression], tokenize_functions: list
    ):
        super().__init__(regular_expressions, tokenize_functions)
        self.dfa_list = [DFA(re) for re in regular_expressions]

    def get_successful_id(self, derivatives: list[_RegularExpression]):
        for index in range(len(derivatives)):
            if derivatives[index].nullable().is_epsilon():
                return index
        return True

    def is_failure_state(self, derivatives: list[_RegularExpression]):
        for derivative_ in derivatives:
            if not derivative_.simplify().is_empty_set():
                return False
        return True

    def find_records(self):
        i = self.i
        j = self.i
        records = []
        derivatives: list[_RegularExpression] = self.regular_expressions.copy()
        while len(self.source_program) > j:
            symbol = self.source_program[j]
            for i_ in range(len(derivatives)):
                derivatives[i_] = self.dfa_list[i_].transition(derivatives[i_], symbol)
            j = j + 1
            records += [[i, j, derivatives.copy()]]
            if self.is_failure_state(derivatives):
                break
        return records



from dataclasses import dataclass
from typing import TypeVar, Generic, Optional, Callable

T = TypeVar("T")


@dataclass(frozen=True)
class Token(Generic[T]):
    identifier: str
    lexeme: str
    value: T


def build_token_func(
    identifier: str, process_lexeme_func: Optional[Callable[[str], T]] = None
) -> Callable[[str], Token[T]]:
    if process_lexeme_func:
        return lambda lexeme: Token(identifier, lexeme, process_lexeme_func(lexeme))
    return lambda lexeme: Token(identifier, lexeme, None)


class Lexer:
    def __init__(self, regular_expressions: list, tokenize_functions: list):
        self.regular_expressions = regular_expressions
        self.tokenize_functions = tokenize_functions
        self.source_program = ""
        if len(regular_expressions) != len(tokenize_functions):
            raise SyntaxError(
                f"Regular Expressions{len(regular_expressions)} and "
                f"Tokenizeing functions{len(tokenize_functions)} are "
                f"not the same length"
            )

    def set_source_program(self, source_program) -> iter:
        self.source_program = source_program
        return iter(self)

    def find_records(self) -> list:
        return []

    def get_successful_id(self, states) -> int:
        return -1

    def is_failure_state(self, derivatives) -> bool:
        return True

    def __iter__(self, **kwargs):
        self.i = 0
        return self

    def __next__(self):
        while True:
            if len(self.source_program) <= self.i:
                raise StopIteration
            records = self.find_records()
            for record in records[::-1]:
                if not self.is_failure_state(record[2]):
                    lexeme = self.source_program[record[0] : record[1]]
                    token_function_id = self.get_successful_id(record[2])
                    self.i = record[1]
                    token = self.tokenize_functions[token_function_id](lexeme)
                    if token is not None:
                        return token
                    else:
                        break

    def peek(self, n=1):
        original_i = self.i
        while True:
            if len(self.source_program) <= self.i:
                return None
            records = self.find_records()
            for record in records[::-1]:
                if not self.is_failure_state(record[2]):
                    lexeme = self.source_program[record[0] : record[1]]
                    token_function_id = self.get_successful_id(record[2])
                    token = self.tokenize_functions[token_function_id](lexeme)
                    if token is not None:
                        n = n - 1
                        if n == 0:
                            self.i = original_i
                            return token
                        else:
                            self.i = record[1]
                            break
                    else:
                        self.i = record[1]
                        break



from string import printable

from pylexers.RegularExpressions.BaseRegularExpressions import (
    _EmptySet,
    _Epsilon,
    _Symbol,
    _Or,
    _Concat,
    _Star,
)

from pylexers.RegularExpressions.BaseRegularExpressions import _RegularExpression

"""
Basic Regular Expressions
"""


class EmptySet(_EmptySet):
    pass


class Epsilon(_Epsilon):
    pass


class Symbol(_Symbol):
    pass


def Or(*regular_expressions: _RegularExpression) -> _RegularExpression:
    re = regular_expressions[0]
    for r in regular_expressions[1:]:
        re = _Or(re, r)
    return re


def Concat(
    regular_expression_1: _RegularExpression, regular_expression_2: _RegularExpression
) -> _RegularExpression:
    return _Concat(regular_expression_1, regular_expression_2)


class Star(_Star):
    pass


"""
Extended Regular Expressions
"""


def Sigma(alphabet: str = printable, exclude: str = "") -> _RegularExpression:
    return Or(*[_Symbol(a) for a in alphabet if a not in exclude])


def String(string: str) -> _RegularExpression:
    if len(string) == 1:
        return _Symbol(string)
    return Concat(_Symbol(string[0]), String(string))


def AtLeastOne(regular_expression: _RegularExpression) -> _RegularExpression:
    return Concat(regular_expression, _Star(regular_expression))


def Optional(regular_expression: _RegularExpression) -> _RegularExpression:
    return _Or(regular_expression, _Epsilon())



from pylexers.BaseLexer import Lexer
from pylexers.NFALexer.NFA import NFA, _combine_nfas
from pylexers.NFALexer.nfa_from_re import _regular_expression_to_nfa


class NFALexer(Lexer):
    def __init__(self, regular_expressions: list, tokenize_functions: list):
        super().__init__(regular_expressions, tokenize_functions)
        nfa_list = [_regular_expression_to_nfa(re) for re in regular_expressions]
        self.tokenize_functions = dict()
        for i in range(len(nfa_list)):
            for accepting_state in nfa_list[i].final_states:
                self.tokenize_functions[accepting_state] = tokenize_functions[i]

        self.initial_state = _combine_nfas(nfa_list)
        self.final_states = self.initial_state.final_states
        self.initial_state.subset_construction()
        self.source_program = ""

    def get_successful_id(self, states):
        for state in states:
            if state.is_final_state(self.final_states):
                return state.is_final_state(self.final_states)
        return False

    def is_failure_state(self, states):
        if states == set():
            return True
        return False

    def find_records(self):
        i = self.i
        j = self.i
        records = []
        states = {self.initial_state}
        while len(self.source_program) > j:
            next_states = set()
            for state in states:
                next_states = next_states.union(state.delta(self.source_program[j]))
            states = next_states
            j += 1
            records += [(i, j, states)]
            if self.is_failure_state(states):
                break
        return records



from pylexers.RegularExpressions import _RegularExpression
from pylexers.RegularExpressions.BaseRegularExpressions import _EmptySet


class DFA:
    def __init__(self, regular_expression: _RegularExpression):
        alphabet = regular_expression.get_alphabet()
        self.states: dict[_RegularExpression, dict] = dict()
        queue = {regular_expression}
        while len(queue) > 0:
            current = queue.pop()
            if current in self.states:
                continue
            self.states[current] = dict()
            for symbol in alphabet:
                self.states[current][symbol] = current.derivative(symbol).simplify()
                queue.add(self.states[current][symbol])

        if _EmptySet() not in self.states.keys():
            self.states[_EmptySet()] = dict()

    def transition(self, state: _RegularExpression, symbol: chr):
        if state not in self.states:
            raise ValueError("Unknown State given to DFA")
        if symbol not in self.states[state]:
            return _EmptySet()
        return self.states[state][symbol]



class _NFA:
    pass


class NFA(_NFA):
    STATE_ID = -1

    def __init__(self, final_states=None, name=""):
        if final_states is None:
            final_states = set()
        NFA.STATE_ID += 1

        self.id = NFA.STATE_ID
        self.final_states = final_states
        self.transitions = {}
        self.name = f"STATE: {self.id}" if name == "" else name

    def add_transition(self, state: _NFA, symbol: str):
        if symbol in self.transitions:
            self.transitions[symbol].add(state)
        else:
            self.transitions[symbol] = {state}

    def subset_construction(self, visited=None):
        if visited is None:
            visited = set()
        if self in visited:
            return
        visited.add(self)

        if "EPSILON" in self.transitions:
            epsilon_transitions = self.transitions["EPSILON"].copy()
            epsilon_visited = set()
            while len(epsilon_transitions) > 0:
                epsilon_state = epsilon_transitions.pop()
                if epsilon_state not in epsilon_visited:
                    epsilon_visited.add(epsilon_state)
                    if "EPSILON" in epsilon_state.transitions:
                        epsilon_transitions = epsilon_transitions.union(
                            epsilon_state.transitions["EPSILON"]
                        )
                    for key in epsilon_state.transitions.keys():
                        for state in epsilon_state.transitions[key]:
                            self.add_transition(state, key)

        for symbol, states in self.transitions.items():
            for state in states:
                state.subset_construction(visited)

    def is_final_state(self, final_states):
        if self in final_states:
            return self
        for state in final_states:
            if "EPSILON" in self.transitions and state in self.transitions["EPSILON"]:
                return state
        return False

    def delta(self, symbol):
        transitions = set()
        if symbol in self.transitions:
            transitions = self.transitions[symbol]
        return transitions

    def __str__(self):
        return self.name + f"({len(self.transitions)})"

    def __repr__(self):
        return str(self)

    def display(self, visited=None):
        if visited is None:
            visited = set()
        if self in visited:
            return
        visited.add(self)
        for symbol in self.transitions:
            for state in self.transitions[symbol]:
                print(f"{self} -- {symbol} --> {state}")

        for symbol in self.transitions:
            for state in self.transitions[symbol]:
                state.display(visited)


def _combine_nfas(nfas: list):
    initial_state = NFA()
    for nfa in nfas:
        initial_state.add_transition(nfa, "EPSILON")
        initial_state.final_states = initial_state.final_states.union(nfa.final_states)
    return initial_state



from setuptools import setup, find_packages

VERSION = '3'
DESCRIPTION = 'PyLexers is a python lexical analyzer generator.'
LONG_DESCRIPTION = 'PyLexer can be used to make many different types of ' \
                   'lexical analyzer generators for your project.'

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="pylexers",
    version=VERSION,
    author="Andrew Reed",
    author_email="andrewreed2017@icloud.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    keywords=['python', 'lexical analysis', 'lexical analyzer generator', 'lexer'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


from pylexers.BaseLexer import Lexer
from pylexers.RegularExpressions.BaseRegularExpressions import _RegularExpression


class DerivativeLexer(Lexer):
    def get_successful_id(self, derivatives: list[_RegularExpression]):
        for index in range(len(derivatives)):
            if derivatives[index].nullable().is_epsilon():
                return index
        return True

    def is_failure_state(self, derivatives: list[_RegularExpression]):
        for derivative_ in derivatives:
            if not derivative_.simplify().is_empty_set():
                return False
        return True

    def find_records(self):
        i = self.i
        j = self.i
        records = []
        derivatives: list[_RegularExpression] = self.regular_expressions.copy()
        while len(self.source_program) > j:
            symbol = self.source_program[j]
            derivatives = [re.derivative(symbol) for re in derivatives]
            j = j + 1
            records += [[i, j, derivatives.copy()]]
            if self.is_failure_state(derivatives):
                break
        return records



from dataclasses import dataclass
from typing import Optional, Union


@dataclass(frozen=True)
class _RegularExpression:
    type: str
    re1: Union["_RegularExpression", str, None] = None
    re2: Optional["_RegularExpression"] = None

    def is_empty_set(self) -> bool:
        return self.type == "EMPTY SET"

    def is_epsilon(self) -> bool:
        return self.type == "EPSILON"

    def is_symbol(self) -> bool:
        return self.type == "SYMBOL"

    def is_concat(self) -> bool:
        return self.type == "CONCAT"

    def is_or(self) -> bool:
        return self.type == "OR"

    def is_kleene_star(self) -> bool:
        return self.type == "KLEENE STAR"

    def is_nullable(self):
        return self.nullable().is_epsilon()

    def nullable(self) -> "_RegularExpression":
        if self.is_epsilon() or self.is_kleene_star():
            return _Epsilon()
        elif self.is_empty_set() or self.is_symbol():
            return _EmptySet()
        elif self.is_concat():
            a = self.re1.nullable()
            b = self.re2.nullable()
            if a.is_epsilon() and b.is_epsilon():
                return _Epsilon()
            return _EmptySet()
        elif self.is_or():
            a = self.re1.nullable()
            b = self.re2.nullable()
            if a.is_epsilon() or b.is_epsilon():
                return _Epsilon()
            return _EmptySet()

    def simplify(self) -> "_RegularExpression":
        if self.is_concat():
            r = self.re1.simplify()
            s = self.re2.simplify()
            if r.is_empty_set() or s.is_empty_set():
                return _EmptySet()
            if r.is_epsilon():
                return s
            if s.is_epsilon():
                return r
        elif self.is_or():
            r = self.re1.simplify()
            s = self.re2.simplify()
            if r.is_empty_set():
                return s
            elif s.is_empty_set():
                return r
        return self

    def derivative(self, symbol) -> "_RegularExpression":
        if self.is_empty_set() or self.is_epsilon():
            return _EmptySet()
        elif self.is_symbol():
            if self.re1 == symbol:
                return _Epsilon()
            else:
                return _EmptySet()
        elif self.is_concat():
            dr = self.re1.derivative(symbol)
            ds = self.re2.derivative(symbol)
            return _Or(
                _Concat(dr, self.re2), _Concat(self.re1.nullable(), ds)
            ).simplify()
        elif self.is_or():
            dr = self.re1.derivative(symbol)
            ds = self.re2.derivative(symbol)
            return _Or(dr, ds).simplify()
        elif self.is_kleene_star():
            dr = self.re1.derivative(symbol)
            return _Concat(dr, self).simplify()
        else:
            raise SyntaxError("Encountered unknown regular expession")

    def matches(self, string):
        re = self
        for c in string:
            re = re.derivative(c).simplify()
        return re.is_nullable()

    def get_alphabet(self) -> str:
        if self.re1 and self.re2:
            return self.re1.get_alphabet() + self.re2.get_alphabet()
        elif isinstance(self.re1, _RegularExpression):
            return self.re1.get_alphabet()
        else:
            return self.re1 or ""


class _EmptySet(_RegularExpression):
    def __init__(self):
        super().__init__("EMPTY SET")


class _Epsilon(_RegularExpression):
    def __init__(self):
        super().__init__("EPSILON")


class _Symbol(_RegularExpression):
    def __init__(self, symbol: str):
        super().__init__("SYMBOL", symbol)


class _Concat(_RegularExpression):
    def __init__(
        self,
        regular_expression_1: _RegularExpression,
        regular_expression_2: _RegularExpression,
    ):
        super().__init__("CONCAT", regular_expression_1, regular_expression_2)


class _Or(_RegularExpression):
    def __init__(
        self,
        regular_expression_1: _RegularExpression,
        regular_expression_2: _RegularExpression,
    ):
        super().__init__("OR", regular_expression_1, regular_expression_2)


class _Star(_RegularExpression):
    def __init__(
        self,
        regular_expression: _RegularExpression,
    ):
        super().__init__("KLEENE STAR", regular_expression)


def _get_regular_expression_arg1(re: _RegularExpression) -> _RegularExpression:
    return re.re1 or _EmptySet()


def _get_regular_expression_arg2(re: _RegularExpression) -> _RegularExpression:
    return re.re2 or _EmptySet()



from pylexers.RegularExpressions.BaseRegularExpressions import (
    _get_regular_expression_arg1,
    _get_regular_expression_arg2,
)
from pylexers.NFALexer.NFA import NFA


def _nfa_from_re_epsilon():
    return NFA()


def _nfa_from_re_symbol(re):
    final_state = NFA()
    initial_state = NFA(final_states={final_state})
    initial_state.add_transition(final_state, _get_regular_expression_arg1(re))
    return initial_state


def _nfa_from_re_concat(re):
    machine_1 = _regular_expression_to_nfa(_get_regular_expression_arg1(re))
    machine_2 = _regular_expression_to_nfa(_get_regular_expression_arg2(re))
    accepting_state = NFA()
    initial_state = NFA(final_states={accepting_state})

    initial_state.add_transition(machine_1, "EPSILON")
    for final_state in machine_1.final_states:
        final_state.add_transition(machine_2, "EPSILON")
    for final_state in machine_2.final_states:
        final_state.add_transition(accepting_state, "EPSILON")
    return initial_state


def _nfa_from_re_or(re):
    machine_1 = _regular_expression_to_nfa(_get_regular_expression_arg1(re))
    machine_2 = _regular_expression_to_nfa(_get_regular_expression_arg2(re))
    accepting_state = NFA()
    initial_state = NFA(final_states={accepting_state})

    initial_state.add_transition(machine_1, "EPSILON")
    initial_state.add_transition(machine_2, "EPSILON")
    for final_state in machine_1.final_states.union(machine_2.final_states):
        final_state.add_transition(accepting_state, "EPSILON")
    return initial_state


def _nfa_from_re_kleene_star(re):
    machine = _regular_expression_to_nfa(_get_regular_expression_arg1(re))
    accepting_state = NFA()
    initial_state = NFA(final_states={accepting_state})

    initial_state.add_transition(machine, "EPSILON")
    initial_state.add_transition(accepting_state, "EPSILON")
    for final_state in machine.final_states:
        final_state.add_transition(machine, "EPSILON")
        final_state.add_transition(accepting_state, "EPSILON")
    return initial_state


def _regular_expression_to_nfa(re):
    if re.is_epsilon():
        return _nfa_from_re_epsilon()
    elif re.is_symbol():
        return _nfa_from_re_symbol(re)
    elif re.is_concat():
        return _nfa_from_re_concat(re)
    elif re.is_or(re):
        return _nfa_from_re_or(re)
    elif re.is_kleene_star():
        return _nfa_from_re_kleene_star(re)
    else:
        raise SyntaxError("Encountered unknown regular expession")
